#include <ode/ode.h>
#include "common.h"

// 以下のexternはmain.hの中で宣言
extern MyObject base;
extern dJointID joint[NUM];
extern dReal WHEEL_BIG_R, WHEEL_W;  // 車輪の半径,幅
extern dReal BASE_R;                // 台車の半径

/*** ロボットの針路計算 ***/
dReal heading()
{
    const dReal *rot = dBodyGetRotation(base.body);
    dReal r11,r12,r13,r21,r22,r23,r31,r32,r33;
    dReal pitch, yaw, roll;

    r11 = rot[0];
    r12 = rot[1];
    r13 = rot[2];
    r21 = rot[4];
    r22 = rot[5];
    r23 = rot[6];
    r31 = rot[8];
    r32 = rot[9];
    r33 = rot[10];
    // ポインタ的表記では　r11 = *(rot+0); r12 = *(rot+1); r13 = *(rot+2);
    pitch = atan2(-r31, sqrt(r32*r32+r33*r33));
    yaw   = atan2(r21,r11);
    roll  = atan2(r32,r33);
    //printf("roll=%f pitch=%f yaw=%f \n", 180.0*roll/M_PI, 180.0*pitch/M_PI, 180.0*yaw/M_PI);
    return -yaw;  // rotation axis is reverse
}


/*** センサ ***/
// 共分散行列Σの誤差分布を持つGPSデータあるいはカメラによる自己位置をシミュレートする関数
void getPos(double sensor_pos[2], double sensor_course)
{
    double cov[3];  // 共分散行列

    // 共分散行列に値を入れる
    cov[0] =  0.5;  // [m]   1行目
    cov[1] =  0.5;  // [m]   ２行目
    cov[2] =  0.1;  // [rad] ３行目

    const double *pb = dBodyGetPosition(base.body);

    sensor_pos[0] = pb[0]; // pbは真の自己位置
    sensor_pos[1] = pb[1];
    sensor_course = heading();

    double r1, r2, r3;
    double m = 0.0;     // 平均

    //ボックス・ミュラー法
    r1 = (double)rand()/RAND_MAX;
    r2 = (double)rand()/RAND_MAX;
    r3 = (double)rand()/RAND_MAX;
    sensor_pos[0] += cov[0]*sqrt(-2*log(r1))*cos(2*M_PI*r2)+m;
    sensor_pos[1] += cov[1]*sqrt(-2*log(r1))*sin(2*M_PI*r2)+m;
    sensor_course += cov[2]*sqrt(-2*log(r1))*cos(2*M_PI*r3)+m;
}



/** 真の自己位置pos[2]と姿勢(ロボットの進行方向）courseを求める関数　**/
void getRealPos(double pos[2], double *course)
{
    const double *pb = dBodyGetPosition(base.body);

    pos[0] = pb[0]; // pbは真の自己位置
    pos[1] = pb[1];
    *course  = heading();
}

// デッドレコニング
void deadReckoning()
{
    static dReal pose = 0, pose_old = 0; // ロボットの針路
    static dReal s_x = 0.0, s_y = 0.0, s_theta = 0.0;   // デッドレコニングの距離,姿勢
    static dReal omega_r = 0, omega_l = 0, omega_r_old = 0, omega_l_old = 0;
    const dReal *pb;
    dReal r, phi; // r:物体までの距離, phi:物体の方位
    dReal edges_x, edges_y, edges_theta;

    // 台車重心の位置
    pb  = dBodyGetPosition(base.body);

    pose    = heading(); // ロボットの針路
    printf("pose=%f \n",pose);
    omega_r = dJointGetHingeAngleRate(joint[RIGHT]);
    omega_l = dJointGetHingeAngleRate(joint[LEFT]);

    // 位置x座標
    edges_x = WHEEL_BIG_R * (omega_r_old + omega_l_old) * sin(pose_old)/2.0
              + WHEEL_BIG_R * (omega_r     + omega_l    ) * sin(pose)/2.0;
    s_x     += 0.5 * STEP_SIZE * edges_x;

    // 位置y座標
    edges_y = WHEEL_BIG_R * (omega_r_old + omega_l_old) * cos(pose_old)/2.0
              + WHEEL_BIG_R * (omega_r     + omega_l    ) * cos(pose)/2.0;
    s_y     += 0.5 * STEP_SIZE * edges_y;

    // 姿勢 theta
    //edges_theta = WH_R1 * (omega_r_old - omega_l_old) /(2.0 * ((2.0 * BASE_R+WHEEL_W)/2.0))
    //			   	  + WH_R1 * (omega_r     - omega_l    ) /(2.0 * ((2.0 * BASE_R+WHEEL_W)/2.0));
    edges_theta = WHEEL_BIG_R * (omega_r_old - omega_l_old) /(2.0 * (BASE_R+WHEEL_W))
                  + WHEEL_BIG_R * (omega_r     - omega_l    ) /(2.0 * (BASE_R+WHEEL_W));
    s_theta     += 0.5 * STEP_SIZE * edges_theta;
    if (s_theta >   M_PI)  s_theta -= 2.0 * M_PI;
    if (s_theta < - M_PI)  s_theta += 2.0 * M_PI;
    //s_theta = - s_theta;  // ロボット姿勢の取り方は逆

    printf("Real:x,y,theta       = (%9.6f,%9.6f,%9.6f) \n", pb[0],pb[1],pose*180.0/M_PI);
    printf("Reckoning: x,y,theta = (%9.6f,%9.6f,%9.6f) \n", s_x,s_y,-s_theta*180.0/M_PI);
    printf("Error = (%f,%f,%f)\n", pb[0]-s_x,pb[1]-s_y,(pose+s_theta)*180.0/M_PI);

    omega_r_old = omega_r;
    omega_l_old = omega_l;
    pose_old    = pose;
}

